import os
import time
import csv
import mysql.connector
import pysftp
from google.cloud import bigquery
from google.cloud import storage
from oauth2client.client import GoogleCredentials
from googleapiclient import discovery
from google.cloud import exceptions


class Setup:

    def __init__(self):

        self.project = os.environ.get('PROJECT', None)
        self.dataset_id = os.environ.get('DATASET_ID', None)
        self.table_id = os.environ.get('TABLE_ID', None)
        self.bucket_name = os.environ.get('BUCKET_NAME', None)
        self.bucket_blob_name = os.environ.get('BUCKET_BLOB_NAME', None)
        self.mysql_instance_name = os.environ.get('MYSQL_INSTANCE_NAME', None)
        self.mysql_username = os.environ.get('MYSQL_USERNAME', None)
        self.mysql_password = os.environ.get('MYSQL_PASSWORD', None)
        self.mysql_database = os.environ.get('MYSQL_DATABASE', None)
        self.mysql_ip_address = os.environ.get('MYSQL_IP_ADDRESS', None)
        self.mysql_port = os.environ.get('MYSQL_PORT', None)
        self.ftp_instance_name = os.environ.get('FTP_INSTANCE_NAME', None)
        self.ftp_hostname = os.environ.get('FTP_HOSTNAME', None)
        self.ftp_username = os.environ.get('FTP_USERNAME', None)
        self.ftp_password = os.environ.get('FTP_PASSWORD', None)
        self.ftp_port = int(os.environ.get('FTP_PORT', 0))
        self.zone = os.environ.get('ZONE', None)

    def create_bq_table(self):
        dataset_ref = bigquery.DatasetReference(project=self.project, dataset_id=self.dataset_id)
        table_ref = bigquery.TableReference(dataset_ref, table_id=self.table_id)
        bigquery_client = bigquery.Client(project=self.project)

        try:
            bigquery_client.get_dataset(dataset_ref)
            self.delete_bq_dataset()
        except exceptions.NotFound as error:
            pass
        finally:
            bigquery_client.create_dataset(dataset_ref)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1,
            autodetect=True,
            field_delimiter=','
        )

        with open('tests/data/sample.csv', "rb") as source_file:
            job = bigquery_client.load_table_from_file(source_file, table_ref, job_config=job_config)
        job.result()

    def delete_bq_dataset(self):
        bigquery_client = bigquery.Client(project=self.project)
        bigquery_client.delete_dataset(self.dataset_id, delete_contents=True)

    def create_bucket(self):
        storage_client = storage.Client(project=self.project)

        try:
            storage_client.get_bucket(self.bucket_name)
            self.delete_bucket()
        except exceptions.NotFound:
            pass
        finally:
            bucket = storage_client.create_bucket(self.bucket_name, location='EU')

        blob = bucket.blob(self.bucket_blob_name)
        blob.upload_from_filename('tests/data/sample.csv')

    def delete_bucket(self):

        storage_client = storage.Client(project=self.project)
        bucket = storage_client.get_bucket(self.bucket_name)

        blobs = storage_client.list_blobs(self.bucket_name)
        for blob in blobs:
            blob.delete()

        bucket.delete()

    def create_mysql_instance(self):

        credentials = GoogleCredentials.get_application_default()
        service = discovery.build('sqladmin', 'v1beta4', credentials=credentials)

        # make the request body
        create_request_body = {
            "name": self.mysql_instance_name,
            "region": "europe-west1",
            "settings": {
                "tier": "db-n1-standard-1",
                "backupConfiguration": {
                    "binaryLogEnabled": False,
                    "enabled": False
                },
                "ipConfiguration": {
                    "authorizedNetworks": [
                        {
                            "value": "91.103.133.0/24"
                        },
                        {
                            "value": "91.103.134.0/24"
                        },
                        {
                            "value": "195.162.12.0/23"
                        },
                        {
                            "value": "93.186.36.0/27"
                        },
                    ]

                }
            },
            "rootPassword": self.mysql_password
        }

        # send the request and get the operation id
        request = service.instances().insert(
            project=self.project,
            body=create_request_body
        )
        response = request.execute()
        operation_id = response['name']

        # wait until the job is done
        status = 'PENDING'
        while status in ['PENDING', 'RUNNING']:
            request = service.operations().get(project=self.project, operation=operation_id)
            status = request.execute()['status']
            time.sleep(2)  # Avoid to hammer the APIs (100queries per users every 100seconds is the maximum).

        if status != 'DONE':
            raise Exception('Failed to create CloudSQL instance, process status {}'.format(status))

    def get_mysql_instance(self):

        credentials = GoogleCredentials.get_application_default()
        service = discovery.build('sqladmin', 'v1beta4', credentials=credentials)

        # send the request and get the operation id
        request = service.instances().get(
            project=self.project,
            instance=self.mysql_instance_name
        )

        response = request.execute()
        self.mysql_ip_address = response['ipAddresses'][0]['ipAddress']

    def create_mysql_table(self):

        connection = mysql.connector.connect(
            host=self.mysql_ip_address,
            user=self.mysql_username,
            password=self.mysql_password,
            database=self.mysql_database
        )

        create_table_query = """
            CREATE TABLE actors (
                profile_id int,
                first_name varchar(255),
                last_name varchar(255)
            )
        """

        connection.autocommit = True
        cursor = connection.cursor()
        cursor.execute(create_table_query)

        insert_query = "INSERT INTO actors (profile_id,first_name,last_name) VALUES"

        with open('tests/data/sample.csv', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            for row in reader:
                insert_query += "({},{},{}),".format(
                    row['profile_id'],
                    "'{}'".format(row['first_name']) if row['first_name'] else 'NULL',
                    "'{}'".format(row['last_name']) if row['last_name'] else 'NULL'
                )

        cursor.execute(insert_query[:-1])

        cursor.close()
        connection.close()

    def delete_mysql_db(self):

        connection = mysql.connector.connect(
            host=self.mysql_ip_address,
            user=self.mysql_username,
            password=self.mysql_password,
        )

        connection.autocommit = True
        cursor = connection.cursor()
        cursor.execute("DROP DATABASE {}".format(self.mysql_database))
        cursor.close()

    def create_mysql_db(self):

        connection = mysql.connector.connect(
            host=self.mysql_ip_address,
            user=self.mysql_username,
            password=self.mysql_password,
        )

        connection.autocommit = True
        cursor = connection.cursor()

        try:
            cursor.execute("CREATE DATABASE {}".format(self.mysql_database))
        except mysql.connector.errors.DatabaseError as error:
            cursor.execute("DROP DATABASE {}".format(self.mysql_database))
            cursor.execute("CREATE DATABASE {}".format(self.mysql_database))

        cursor.close()
        connection.close()

    def wait_for_operation(self, compute_client, operation):
        while True:
            result = compute_client.zoneOperations().get(
                project=self.project,
                zone=self.zone,
                operation=operation).execute()

            if result['status'] == 'DONE':
                if 'error' in result:
                    raise Exception(result['error'])
                return result

            time.sleep(1)

    def start_sftp_instance(self):

        compute = discovery.build('compute', 'v1')
        operation = compute.instances().start(
            project=self.project,
            zone=self.zone,
            instance=self.ftp_instance_name
        ).execute()
        self.wait_for_operation(compute, operation['name'])
        self.wait_for_instance_startup(compute)
        self.set_sftp_instance_address()

    def wait_for_instance_startup(self, compute_connection):

        status = ''
        while status != 'RUNNING':
            status = compute_connection.instances().get(
                project=self.project,
                zone=self.zone,
                instance=self.ftp_instance_name
            ).execute()['status']

    def set_sftp_instance_address(self):

        compute = discovery.build('compute', 'v1')
        instance_details = compute.instances().get(project='tmg-plat-dev', zone='europe-west2-c', instance='library-sftp-tests-instance').execute()
        try:
            self.ftp_hostname = instance_details['networkInterfaces'][0]['accessConfigs'][0]['natIP']
        except IndexError as e:
            print("failed lookup of sftp test instance address")
            raise e
        except KeyError as e:
            print("failed lookup of sftp test instance address")
            raise e

    def upload_ftp_files(self):
        needed_remote_files = [
            "list_sample1.csv",
            "list_sample2.csv",
            "list_sample3.csv",
            "test_dir/test_subdir/list_sample1.csv",
            "test_dir/test_subdir/list_sample2.csv",
            "test_dir/test_subdir/list_sample3.csv",
            "test_dir/download_sample.csv",
            "test_dir/permission_change.csv",
            "test_dir/permission_change2.csv",
            "test_dir/delete_sample.csv"
        ]

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        connection = pysftp.Connection(
            host=self.ftp_hostname,
            username=self.ftp_username,
            password=self.ftp_password,
            port=self.ftp_port,
            cnopts=cnopts
        )

        connection.makedirs("test_dir/test_subdir")

        for file in needed_remote_files:
            if not connection.exists(file):
                connection.put('tests/data/sample.csv', file)

    def remove_ftp_files(self):
        unwanted_remote_files = [
            "sample.csv",
            "test_dir/sample.csv",
            "test_dir/test_subdir/rename_sample.csv",
            "test_dir/bq_to_ftp_test_sample.csv"
        ]

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None

        connection = pysftp.Connection(
            host=self.ftp_hostname,
            username=self.ftp_username,
            password=self.ftp_password,
            port=self.ftp_port,
            cnopts=cnopts
        )

        for file in unwanted_remote_files:
            if connection.exists(file):
                connection.remove(file)

        unwanted_local_files = [
            "download_sample.csv",
            "tests/data/download_sample.csv",
            "tests/data/rename_sample.csv",
            "tests/data/download_sample.csv",
            "tests/data/bq_to_ftp_test_sample.csv"
        ]

        for file in unwanted_local_files:
            if os.path.exists(file):
                os.remove(file)

    def stop_sftp_instance(self):

        compute = discovery.build('compute', 'v1')
        operation = compute.instances().stop(
            project=self.project,
            zone=self.zone,
            instance=self.ftp_instance_name
        ).execute()
        self.wait_for_operation(compute, operation['name'])



setup = Setup()



import unittest
import csv
import pysftp
import mysql.connector
from google.cloud import bigquery
from google.cloud import storage

from tmg.data import transfer
from tests.setup import setup


def setUpModule():
    setup.create_bq_table()
    setup.create_bucket()
    setup.create_mysql_db()
    setup.create_mysql_table()
    setup.start_sftp_instance()
    setup.upload_ftp_files()


def tearDownModule():
    setup.delete_bq_dataset()
    setup.delete_bucket()
    setup.delete_mysql_db()
    setup.remove_ftp_files()
    setup.stop_sftp_instance()


class TestTransfer(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestTransfer, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_bq_to_mysql(self):

        # create the actors_from_bq table in mysql
        connection = mysql.connector.connect(
            host=self.setup.mysql_ip_address,
            user=self.setup.mysql_username,
            password=self.setup.mysql_password,
            database=self.setup.mysql_database
        )
        create_table_query = """
            CREATE TABLE actors_from_bq (
                profile_id int,
                first_name varchar(255),
                last_name varchar(255)
            )
        """
        connection.autocommit = True
        cursor = connection.cursor()
        cursor.execute(create_table_query)

        # run the transfer
        client = transfer.Client(project=self.setup.project)
        client.bq_to_mysql(
            connection_string='{username}:{password}@{host}:{port}/{database}'.format(
                username=self.setup.mysql_username,
                password=self.setup.mysql_password,
                host=self.setup.mysql_ip_address,
                port=self.setup.mysql_port,
                database=self.setup.mysql_database
            ),
            bq_table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, self.setup.table_id),
            mysql_table='{}.{}'.format(self.setup.mysql_database, 'actors_from_bq')
        )

        # creating list based on mysql exported table values
        cursor.execute('SELECT profile_id, first_name, last_name from actors_from_bq')
        mysql_keys = []
        for row in cursor.fetchall():
            mysql_keys.append(
                str(row[0]) +
                'NULL' if row[1] is None else row[1] +
                'NULL' if row[2] is None else row[2]
            )
        cursor.close()
        connection.close()

        # creating list based on source big query table values
        bq_keys = []
        biquery_client = bigquery.Client(project=self.setup.project)
        job = biquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(self.setup.dataset_id, self.setup.table_id)
        )
        for row in job.result():
            bq_keys.append(
                str(row.profile_id) +
                'NULL' if row.first_name is None else row.first_name +
                'NULL' if row.last_name is None else row.last_name
            )

        self.assertEqual(mysql_keys, bq_keys)

    def test_mysql_to_bq(self):

        # run the transfer
        client = transfer.Client(project=self.setup.project)
        client.mysql_to_bq(
            instance_name=self.setup.mysql_instance_name,
            database=self.setup.mysql_database,
            query='SELECT * from actors',
            bq_table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'actors_from_mysql'),
            bq_table_schema=(('profile_id', 'STRING'), ('first_name', 'STRING'), ('last_name', 'STRING'))
        )

        # creating list based on source mysql table values
        connection = mysql.connector.connect(
            host=self.setup.mysql_ip_address,
            user=self.setup.mysql_username,
            password=self.setup.mysql_password,
            database=self.setup.mysql_database
        )
        cursor = connection.cursor()
        cursor.execute('SELECT profile_id, first_name, last_name from actors')
        mysql_keys = []
        for row in cursor.fetchall():
            mysql_keys.append(
                str(row[0]) +
                'NULL' if row[1] is None else row[1] +
                'NULL' if row[2] is None else row[2]
            )
        cursor.close()
        connection.close()

        # creating list based on exported big query table values
        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'actors_from_mysql'
            )
        )
        for row in job.result():
            bq_keys.append(
                str(row.profile_id) +
                'NULL' if row.first_name is None else row.first_name +
                'NULL' if row.last_name is None else row.last_name
            )

        self.assertEqual(mysql_keys, bq_keys)

    def test_bq_to_gs(self):

        client = transfer.Client(project=self.setup.project)
        file_names = client.bq_to_gs(
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'actors'),
            bucket_name=self.setup.bucket_name,
        )

        # creating list based on source big query table values
        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'actors'
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        # creating list based on destination bucket file values
        storage_client = storage.Client(project=self.setup.project)
        with open('actors.csv', 'wb') as file_obj:
            storage_client.download_blob_to_file(blob_or_uri=file_names[0], file_obj=file_obj)

        with open('actors.csv', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append(
                    str(row['profile_id']) + row['first_name'] + row['last_name']
                )

        self.assertEqual(storage_keys, bq_keys)

    def test_mysql_to_gs(self):

        client = transfer.Client(project=self.setup.project)
        client.mysql_to_gs(
            instance_name=self.setup.mysql_instance_name,
            database=self.setup.mysql_database,
            query="SELECT * FROM actors",
            gs_uri="gs://{}/actors_from_my_sql.csv".format(self.setup.bucket_name)
        )

        # creating list based on source mysql table values
        connection = mysql.connector.connect(
            host=self.setup.mysql_ip_address,
            user=self.setup.mysql_username,
            password=self.setup.mysql_password,
            database=self.setup.mysql_database
        )
        cursor = connection.cursor()
        cursor.execute('SELECT profile_id, first_name, last_name from actors')
        mysql_keys = []
        for row in cursor.fetchall():
            mysql_keys.append('{}{}{}'.format(
                row[0],
                'NULL' if row[1] is None else row[1],
                'NULL' if row[2] is None else row[2]
            ))
        cursor.close()
        connection.close()

        # creating list based on destination bucket values
        storage_client = storage.Client(project=self.setup.project)
        with open('actors_from_my_sql.csv', 'wb') as file_obj:
            storage_client.download_blob_to_file(
                blob_or_uri="gs://{}/actors_from_my_sql.csv".format(self.setup.bucket_name),
                file_obj=file_obj
            )
        with open('actors_from_my_sql.csv', newline='') as csvfile:
            reader = csv.reader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append(
                    "{}{}{}".format(
                        row[0],
                        "NULL" if row[1] == 'N\n' or row[1] == 'N' else row[1],
                        "NULL" if row[2] == 'N\n' or row[2] == 'N' else row[2],
                    )
                )

        self.assertEqual(storage_keys, mysql_keys)

    def test_gs_to_bq(self):

        client = transfer.Client(project=self.setup.project)
        client.gs_to_bq(
            gs_uri="gs://{}/{}".format(self.setup.bucket_name,self.setup.bucket_blob_name),
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'actors_from_gs')
        )

        # creating list based on source bucket file values
        storage_client = storage.Client(project=self.setup.project)
        with open('actors_from_gs.csv', 'wb') as file_obj:
            storage_client.download_blob_to_file(
                blob_or_uri="gs://{}/{}".format(self.setup.bucket_name,self.setup.bucket_blob_name),
                file_obj=file_obj
            )
        with open('actors_from_gs.csv', newline='') as csvfile:
            reader = csv.reader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append(
                    "{}{}{}".format(
                        row[0],
                        "NULL" if row[1] else row[1],
                        "NULL" if row[2] else row[2],
                    )
                )

        # creating list based on destination big query table values
        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'actors_from_gs'
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                'NULL' if row.first_name is None else row.first_name,
                'NULL' if row.last_name is None else row.last_name,
            ))

    def test_ftp_to_bq(self):

        client = transfer.Client(self.setup.project)
        client.ftp_to_bq(
            ftp_connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}",
            bq_table=f"{self.setup.project}.{self.setup.dataset_id}.ftp_test_load",
            ftp_filepath="test_dir/download_sample.csv",
        )

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        connection = pysftp.Connection(
            host=self.setup.ftp_hostname,
            username=self.setup.ftp_username,
            password=self.setup.ftp_password,
            port=self.setup.ftp_port,
            cnopts=cnopts
        )
        connection.get("test_dir/download_sample.csv", 'tests/data/ftp_to_bq_test_sample.csv')
        ftp_keys = []
        with open("tests/data/ftp_to_bq_test_sample.csv", 'rt') as f:
            reader = csv.reader(f)
            reader.__next__()
            for row in reader:
                ftp_keys.append(
                    "{}{}{}".format(
                        row[0],
                        row[1] if row[1] else "NULL",
                        row[2] if row[2] else "NULL",
                    )
                )

        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'ftp_test_load'
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                'NULL' if row.first_name is None else row.first_name,
                'NULL' if row.last_name is None else row.last_name,
            ))

        self.assertListEqual(bq_keys, ftp_keys)

    def test_bq_to_ftp(self):

        cli = transfer.Client(self.setup.project)
        cli.bq_to_ftp(
            ftp_connection_string=f"{self.setup.ftp_username}:{self.setup.ftp_password}@{self.setup.ftp_hostname}:{self.setup.ftp_port}",
            bq_table=f"{self.setup.project}.{self.setup.dataset_id}.{self.setup.table_id}",
            ftp_filepath="test_dir/bq_to_ftp_test_sample.csv",
            separator="|",
            print_header=False
        )

        cnopts = pysftp.CnOpts()
        cnopts.hostkeys = None
        connection = pysftp.Connection(
            host=self.setup.ftp_hostname,
            username=self.setup.ftp_username,
            password=self.setup.ftp_password,
            port=self.setup.ftp_port,
            cnopts=cnopts
        )
        connection.get("test_dir/bq_to_ftp_test_sample.csv", 'tests/data/bq_to_ftp_test_sample.csv')
        ftp_keys = []
        with open("tests/data/bq_to_ftp_test_sample.csv", 'rt') as f:
            reader = csv.reader(f, delimiter="|")
            for row in reader:
                ftp_keys.append(
                    "{}{}{}".format(
                        row[0],
                        "NULL" if row[1] is None else row[1],
                        "NULL" if row[2] is None else row[2],
                    )
                )

        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                self.setup.table_id
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                'NULL' if row.first_name is None else row.first_name,
                'NULL' if row.last_name is None else row.last_name,
            ))



import time
import uuid

from google.cloud import bigquery
from google.cloud import storage
from oauth2client.client import GoogleCredentials
from googleapiclient import discovery

from tmg.data import bq
from tmg.data import mysql
from tmg.data import ftp
from tmg.data._helper import merge_files
from tmg.data._helper import clean_mysql_export, get_bq_write_disposition
from tmg.data import logs



class Client:
    """
    Client to bundle transfers from a source to destination.

    Args:
        project (str): The Project ID for the project which the client acts on behalf of.
    """

    def __init__(self, project):
        self.project = project

    def bq_to_gs(self, table, bucket_name, separator=',', print_header=True):
        """Extract BigQuery table into the GoogleStorage

        Args:
            table (str):  The BigQuery table name. For example: ``my-project-id.you-dataset.my-table``
            bucket_name (str):  The name of the bucket in GoogleStorage.
            separator (:obj:`str`, optional): The separator. Defaults to :data:`,`.
            print_header (:obj:`boolean`, optional):  True to print a header row in the exported data otherwise False. Defaults to :data:`True`.

        Returns:
            list: The list of GoogleStorage paths for the uploaded files into the GoogleStorage.
            if the table is big, the exported files will be multiple.

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_gs('my-project-id.some_dataset.some_table', 'some-bucket-name')
        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(project=project, dataset_id=dataset_id)
        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        bq_client = bigquery.Client(project=self.project)
        logs.client.logger.info('Extracting from {table} to gs://{bucket_name}/{table_id}_*'.format(
            bucket_name=bucket_name, table_id=table_id, table=table)
        )
        extract_job = bq_client.extract_table(
            source=table_ref,
            destination_uris='gs://{bucket_name}/{table_id}_*'.format(bucket_name=bucket_name, table_id=table_id),
            job_config=bigquery.ExtractJobConfig(field_delimiter=separator, print_header=print_header)
        )
        extract_job.result()

        storage_client = storage.Client(project=self.project)
        logs.client.logger.info('Getting the list of available blobs in gs://{}'.format(bucket_name))
        blobs = storage_client.list_blobs(bucket_name)

        return ['gs://{}/{}'.format(bucket_name, blob.name) for blob in blobs]

    def gs_to_bq(self, gs_uri, table, auto_detect=True, skip_leading_rows=True, separator=',',
                 write_preference='append', schema=()):
        """Load file from Google Storage into the BigQuery table

        Args:
            gs_uri (str):  The Google Storage uri for the file. For example: ``gs://my_bucket_name/my_filename``.
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            auto_detect (boolean, Optional):  True if the schema should automatically be detected otherwise False. Defaults to :data:`True`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            separator (str, Optional): The separator. Defaults to :data:`,`
            write_preference (str, Optional): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
                                              Defaults to 'append'

            schema (tuple): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.gs_to_bq(gs_uri='gs://my-bucket-name/my-filename',table='my-project-id.my_dataset.my_table')
        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(project=project, dataset_id=dataset_id)
        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1 if skip_leading_rows else 0,
            autodetect=auto_detect,
            field_delimiter=separator,
            write_disposition=get_bq_write_disposition(write_preference),
            allow_quoted_newlines=True
        )

        if schema:
            job_config.schema = [bigquery.SchemaField(schema_field[0], schema_field[1]) for schema_field in schema]

        bigquery_client = bigquery.Client(project=self.project)
        logs.client.logger.info('Loading BigQuery table {} from {}'.format(table, gs_uri))
        job = bigquery_client.load_table_from_uri(gs_uri, table_ref, job_config=job_config)

        job.result()

    def bq_to_mysql(self, connection_string, bq_table, mysql_table):
        """Export from BigQuery table to MySQL table

        .. note:: The CloudSQL service account which can be found in Cloud SQL UI
              needs to have a storage object admin access.

        Args:
            connection_string (str): The MySQL connection string. For example: ``{my-username}:{my-password}@{mysql-host}:{mysql-port}/{my-database}``
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            mysql_table (str):  The MySQL table. For example: ``my-mysql-database.my-table``

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_mysql(
            >>>     connection_string='{my-username}:{my-password}@{mysql-host}:{mysql-port}/{my-database}'
            >>>     bq_table='my-project-id.my-dataset.my-table'
            >>>     mysql_table='my-mysql-database.my-mysql-table'
            >>> )
        """

        # download the files locally
        bq_client = bq.Client(self.project)
        file_names = bq_client.download_table(bq_table, print_header=False)

        # merge the downloaded files
        if len(file_names) > 1:
            logs.client.logger.info('Merging {}'.format(','.join(file_names)))
            output_file_name = merge_files(file_names)
        else:
            output_file_name = file_names[0]

        # upload the merged file into mysql
        database, table = mysql_table.split(".")
        mysql_client = mysql.Client(connection_string)
        mysql_client.upload_table(file_path=output_file_name, database=database, table=table)

    def mysql_to_gs(self, instance_name, database, query, gs_uri):
        """Export from MySQL to Google Storage in CSV format

        .. note:: Be aware that the exported CSV file has "N as a field value when there is no value for that field.

        Args:
            instance_name (str): The CloudSQL instance name. It's the instance name in CloudSQL UI but without project.location
            database (str): The MySQL database name
            query (str): The query to get the data from MySQL
            gs_uri (str): The GoogleStorage uri. For example: ``gs://bucket_name/file_name``

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.mysql_to_gs(
            >>>     instance_name='{my-mysql-instance-name}',
            >>>     database='my-database',
            >>>     query='SELECT * FROM my-table',
            >>>     gs_uri='gs://my-bucket-name/my-filename'
            >>> )
        """

        # make the discovery service
        credentials = GoogleCredentials.get_application_default()
        service = discovery.build('sqladmin', 'v1beta4', credentials=credentials)

        # make the request body
        instances_export_request_body = {
            "exportContext": {
                "fileType": "CSV",
                "uri": gs_uri,
                "databases": [database],
                "csvExportOptions": {
                    "selectQuery": query
                }
            }
        }

        # send the request and get the operation id
        logs.client.logger.info('Exporting from MySQL database {} with query "{}" to {}'.format(database, query, gs_uri))
        request = service.instances().export(
            project=self.project,
            instance=instance_name,
            body=instances_export_request_body
        )
        response = request.execute()
        operation_id = response['name']

        # wait until the job is done
        status = 'PENDING'
        while status in ['PENDING', 'RUNNING']:
            request = service.operations().get(project=self.project, operation=operation_id)
            status = request.execute()['status']
            time.sleep(2)  # Avoid to hammer the APIs (100queries per users every 100seconds is the maximum).

        if status != 'DONE':
            raise Exception('Failed to export data from MySQL, process status {}'.format(status))

    def mysql_to_bq(self, instance_name, database, query, bq_table, bq_table_schema, write_preference='append'):
        """Export from MySQL to BigQuery

        Args:
            instance_name (str): The CloudSQL instance name. It's the instance name in CloudSQL UI but without project.location
            database (str):  The MySQL database name
            query (str): The query to get the data from MySQL
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            bq_table_schema (tuple): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``
            write_preference (str, Optional): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
                                              Defaults to 'append'

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.mysql_to_bq(
            >>>     instance_name='{my-mysql-instance-name}',
            >>>     database='my-database',
            >>>     query='SELECT * FROM my-table',
            >>>     bq_table='my-project-id.my-dataset.my-table',
            >>>     bq_table_schema=(('first_field','STRING'),('second_field', 'STRING'))
            >>> )

        """

        # make the temp bucket
        storage_client = storage.Client(project=self.project)
        tmp_bucket_name = str(uuid.uuid4())
        logs.client.logger.info('Creating the temporary bucket {}'.format(tmp_bucket_name))
        tmp_bucket = storage_client.create_bucket(tmp_bucket_name, location='EU')
        tmp_file_name = str(uuid.uuid4())

        # export from mysql to temp bucket
        self.mysql_to_gs(
            instance_name=instance_name,
            database=database,
            query=query,
            gs_uri='gs://{}/{}.csv'.format(tmp_bucket_name, tmp_file_name)
        )

        # download the exported file from the temp bucket into local
        blob = tmp_bucket.blob('{}.csv'.format(tmp_file_name))
        logs.client.logger.info('Downloading {}'.format(blob.name))
        blob.download_to_filename('{}.csv'.format(tmp_file_name))

        # clean the file by replacing ,"N with space
        logs.client.logger.info('Cleaning the exported file {}.csv'.format(tmp_file_name))
        cleaned_file_name = clean_mysql_export('{}.csv'.format(tmp_file_name))

        # upload the cleaned file into BigQuery
        bq_client = bq.Client(project=self.project)
        bq_client.upload_table(
            file_path=cleaned_file_name,
            table=bq_table,
            auto_detect=False,
            schema=bq_table_schema,
            skip_leading_rows=False,
            write_preference=write_preference
        )

        # cleanup
        blob.delete()
        logs.client.logger.info('Deleting temporary bucket {}'.format(tmp_bucket_name))
        tmp_bucket.delete()

    def ftp_to_bq(self, ftp_connection_string, ftp_filepath, bq_table, separator=',',
                  skip_leading_rows=True, write_preference='append'):
        """Export from FTP to BigQuery

        Args:
            ftp_connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            ftp_filepath (str): The path to the file to download.
            separator (:obj:`str`, optional): The separator. Defaults to :data:`,`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            write_preference (str, Optional): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
                                              Defaults to 'append'
        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.ftp_to_bq(
            >>>     ftp_connection_string='username:password@hots:port',
            >>>     ftp_filepath='/my-path/to-the-ftp-file',
            >>>     bq_table='my-project-id.my-dataset.my-table'
            >>> )

        """

        # download the ftp file
        ftp_client = ftp.Client(connection_string=ftp_connection_string)
        local_file = ftp_client.download_file(ftp_filepath)

        # upload the ftp file into BigQuery
        bq_cli = bq.Client(project=self.project)
        bq_cli.upload_table(
            file_path=local_file,
            table=bq_table,
            separator=separator,
            skip_leading_rows=skip_leading_rows,
            write_preference=write_preference
        )

    def bq_to_ftp(self, bq_table, ftp_connection_string, ftp_filepath, separator=',', print_header=True):

        """Export from BigQuery to FTP

        Args:
            bq_table (str): The BigQuery table. For example: ``my-project-id.my-dataset.my-table``
            ftp_connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
            ftp_filepath (str): The path to the file to download.
            separator (:obj:`str`, optional): The separator. Defaults to :data:`,`.
            print_header (boolean, Optional):  True to write header for the CSV file, otherwise False. Defaults to :data:`True`.

        Examples:
            >>> from tmg.data import transfer
            >>> client = transfer.Client(project='my-project-id')
            >>> client.bq_to_ftp(
            >>>     bq_table='my-project-id.my-dataset.my-table',
            >>>     ftp_connection_string='username:password@hots:port',
            >>>     ftp_filepath='/my-path/to-the-ftp-file'
            >>> )

        """
        # download the the BigQuery table into local
        bq_client = bq.Client(project=self.project)
        local_files = bq_client.download_table(
            table=bq_table,
            separator=separator,
            print_header=print_header
        )

        # merge the files if they are more than one
        if len(local_files) > 1:
            logs.client.logger.info('Merging {}'.format(','.join(local_files)))
            merged_file = merge_files(local_files)
        else:
            merged_file = local_files[0]

        # upload the merged file
        ftp_client = ftp.Client(connection_string=ftp_connection_string)
        ftp_client.upload_file(local_path=merged_file, remote_path=ftp_filepath)



import uuid
import csv

from google.cloud import storage
from google.cloud import bigquery
from tmg.data import transfer
from tmg.data._helper import get_bq_write_disposition
from tmg.data import logs


class Client:
    """
    Client to bundle BigQuery functionality.

    Args:
        project (str): The Project ID for the project which the client acts on behalf of.
    """

    def __init__(self, project):
        self.project = project

    def download_table(self, table, local_folder='.', separator=',', print_header=True):
        """
        Export the table to the local file in CSV format.

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            local_folder (:obj:`str`, Optional):  The local folder with out the slash at end. For example: ``/some_path/some_inside_path``. Defaults to current path :data:`.`
            separator (:obj:`str`, Optional): The separator. Defaults to :data:`,`
            print_header (:obj:`boolean`, optional):  True to print a header row in the exported file otherwise False. Defaults to :data:`True`.

        Returns:
            list: list of file names, if the table is big, multiple files are downloaded

        Examples:
            >>> from tmg.data import bq
            >>> client = bq.Client(project='my-project-id')
            >>> client.download_table(table='my-project-id.my_dataset.my_table')

        """

        storage_client = storage.Client(project=self.project)
        bucket_name = str(uuid.uuid4())
        logs.client.logger.info('Creating temporary bucket gs://{}'.format(bucket_name))
        tmp_bucket = storage_client.create_bucket(bucket_name, location='EU')

        transfer_client = transfer.Client(project=self.project)
        transfer_client.bq_to_gs(table, bucket_name, separator=separator, print_header=print_header)

        logs.client.logger.info('Getting the list of available blobs in gs://{}'.format(bucket_name))
        blobs = storage_client.list_blobs(bucket_name)
        blob_names = []
        for blob in blobs:
            logs.client.logger.info('Downloading gs://{}/{}'.format(bucket_name, blob.name))
            blob.download_to_filename('{}/{}'.format(local_folder, blob.name))
            blob_names.append(blob.name)
            logs.client.logger.info('Deleting gs://{}/{}'.format(bucket_name, blob.name))
            blob.delete()

        logs.client.logger.info('Deleting bucket gs://{}'.format(bucket_name))
        tmp_bucket.delete()

        return blob_names

    def upload_table(self, file_path, table, separator=',', auto_detect=True, skip_leading_rows=True,
                     write_preference='append', schema=()):
        """Import into the BigQuery table from the local file.

        Args:
            file_path (str):  The local file path
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            separator (str, Optional): The separator. Defaults to :data:`,`
            auto_detect (boolean, Optional):  True if the schema should automatically be detected otherwise False. Defaults to :data:`True`.
            skip_leading_rows (boolean, Optional):  True to skip the first row of the file otherwise False. Defaults to :data:`True`.
            write_preference (str, Optional): The option to specify what action to take when you load data from a source file. Value can be on of
                                              ``'empty'``: Writes the data only if the table is empty.
                                              ``'append'``: Appends the data to the end of the table.
                                              ``'truncate'``: Erases all existing data in a table before writing the new data.
                                              Defaults to 'append'
            schema (tuple, Optional): The BigQuery table schema. For example: ``(('first_field','STRING'),('second_field', 'STRING'))``. Defaults to ().

        Examples:
            >>> from tmg.data import bq
            >>> client = bq.Client(project='my-project-id')
            >>> client.upload_table(file_path='/my_path/my_filename', table='my-project-id.my-dataset.my-table')

        """

        project, dataset_id, table_id = table.split('.')
        dataset_ref = bigquery.DatasetReference(project=project, dataset_id=dataset_id)
        table_ref = bigquery.TableReference(dataset_ref, table_id=table_id)

        job_config = bigquery.LoadJobConfig(
            source_format=bigquery.SourceFormat.CSV,
            skip_leading_rows=1 if skip_leading_rows else 0,
            autodetect=auto_detect,
            field_delimiter=separator,
            write_disposition=get_bq_write_disposition(write_preference),
            allow_quoted_newlines=True
        )

        if schema:
            job_config.schema = [bigquery.SchemaField(schema_field[0], schema_field[1]) for schema_field in schema]

        bigquery_client = bigquery.Client(project=self.project)
        with open(file_path, "rb") as source_file:
            logs.client.logger.info('Loading BigQuery table {} from file {}'.format(table, file_path))
            job = bigquery_client.load_table_from_file(source_file, table_ref, job_config=job_config)

        job.result()

    def run_query(self, query, query_file_name=None, params=(), destination=None):
        """Run the query and return the result

        Args:
            query (str): The query string which be parametrised using ``@``.
                         For example: ``SELECT first_name FROM profile where last_name=@last_name``
                         More info available at https://cloud.google.com/bigquery/docs/parameterized-queries
            query_file_name (str, Optional): The query file name. For example: ``/my_query_path/my_query.sql``
            params (tuple, Optional): The query parameters. Each element should be (param_name, param_type, param_value).
                            For example. ``(("last_name", "STRING", "Deniro"),)``
            destination (str, Optional): The BigQuery destination table name. For example: ``project.dataset.table``. Defaults to None

        Returns:
            The query results

        """

        bigquery_client = bigquery.Client(project=self.project)
        job_config = bigquery.QueryJobConfig(allow_large_results=True)
        if destination:
            job_config.destination = destination
        if params:
            job_config.query_parameters = [
                bigquery.ScalarQueryParameter(param[0], param[1], param[2]) for param in params
            ]
        if query_file_name:
            with open(query_file_name, mode="rb") as query_file:
                query = query_file.read()

        query_job = bigquery_client.query(query, job_config=job_config)
        return query_job.result()

    def create_table(self, table, schema_file_name=None, schema_fields=None):
        """Create the BigQuery table

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
            schema_file_name (str, Optional): The schema file name. For example: ``/my_schema_path/schema.csv``.
                                              The schema file should be in CSV format and each line is FIELD_NAME, FIELD_TYPE, MODE. For example:
                                              ``id,STRING,REQUIRED``. Defaults to None
            schema_fields (tuple, Optional): The table schema.
                                             For example. ``(("id", "INT64", "REQUIRED"),("name", "STRING", "NULLABLE"))``. Defaults to None.

        """

        if not schema_fields and not schema_file_name:
            raise ValueError('Either schema file name or schema fields should be provided.')

        if schema_file_name:
            with open(schema_file_name) as schema_file:
                reader = csv.reader(schema_file)
                schema = [
                    bigquery.SchemaField(row[0], row[1], mode=row[2])
                    for row in reader
                ]
        else:
            schema = [
                bigquery.SchemaField(schema_field[0], schema_field[1], mode=schema_field[2])
                for schema_field in schema_fields
            ]

        table = bigquery.Table(table, schema=schema)
        bigquery_client = bigquery.Client(project=self.project)
        bigquery_client.create_table(table)

    def delete_table(self, table):
        """Delete the BigQuery table

        Args:
            table (str): The BigQuery table name. For example: ``project.dataset.table``.
        """
        bigquery_client = bigquery.Client(project=self.project)
        bigquery_client.delete_table(table)

    def create_dataset(self, dataset, location='EU'):
        """Create the BigQuery dataset

        Args:
            dataset (str): The BigQuery dataset name. For example: ``my-project.my_dataset``.
            location (str, Optional): The location of the dataset. Defaults to EU.
        """
        bigquery_client = bigquery.Client(project=self.project)
        dataset = bigquery.Dataset(dataset)
        dataset.location = location
        bigquery_client.create_dataset(dataset)

    def delete_dataset(self, dataset, delete_contents=False):
        """Delete the BigQuery dataset

        Args:
            dataset (str): The BigQuery dataset name. For example: ``my-project.my_dataset``.
            delete_contents (boolean, Optional): If True, delete all the tables in the dataset. If
                False and the dataset contains tables, the request will fail.
                Default is False.
        """
        bigquery_client = bigquery.Client(project=self.project)
        bigquery_client.delete_dataset(dataset, delete_contents=delete_contents)




import unittest
import csv
from google.cloud import bigquery

from tmg.data import bq
from tests.setup import setup


def setUpModule():
    setup.create_bq_table()
    setup.create_bucket()


def tearDownModule():
    setup.delete_bq_dataset()
    setup.delete_bucket()


class TestBQ(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestBQ, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_download_table(self):

        bq_client = bq.Client(project=self.setup.project)
        downloaded_file_names = bq_client.download_table(
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, self.setup.table_id)
        )

        # creating list based on destination bucket file values
        with open(downloaded_file_names[0], newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append(
                    str(row['profile_id']) + row['first_name'] + row['last_name']
                )

        # creating list based on source big query table values
        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                self.setup.table_id
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        # creating list based on destination bucket file values

        self.assertEqual(storage_keys, bq_keys)

    def test_upload_table(self):

        bq_client = bq.Client(project=self.setup.project)
        bq_client.upload_table(
            file_path='tests/data/sample.csv',
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'uploaded_actors')
        )

        # creating list based on destination bucket file values
        with open('tests/data/sample.csv', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append(
                    str(row['profile_id']) + row['first_name'] + row['last_name']
                )

        # creating list based on source big query table values
        bq_keys = []
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            'SELECT profile_id, first_name, last_name from {}.{}'.format(
                self.setup.dataset_id,
                'uploaded_actors'
            )
        )
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        # creating list based on destination bucket file values

        self.assertEqual(storage_keys, bq_keys)

    def test_run_query(self):

        # creating list based on run_query return's results
        bq_client = bq.Client(project=self.setup.project)
        results = bq_client.run_query(
            query='SELECT * FROM {}.{} where profile_id=@id'.format(self.setup.dataset_id, self.setup.table_id),
            params=(('id', 'INT64', 1),)

        )
        keys = []
        for row in results:
            keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        # creating list based on source big query table values
        bigquery_client = bigquery.Client(project=self.setup.project)
        job = bigquery_client.query(
            query='SELECT profile_id, first_name, last_name from {}.{} where profile_id=@id'.format(
                self.setup.dataset_id,
                self.setup.table_id
            ),
            job_config=bigquery.QueryJobConfig(
                query_parameters=[bigquery.ScalarQueryParameter('id', 'INT64', 1)]
            )
        )
        bq_keys = []
        for row in job.result():
            bq_keys.append('{}{}{}'.format(
                row.profile_id,
                '' if row.first_name is None else row.first_name,
                '' if row.last_name is None else row.last_name,
            ))

        self.assertEqual(keys, bq_keys)

    def test_create_table_with_schema_fields(self):

        bq_client = bq.Client(project=self.setup.project)
        bq_client.create_table(
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'venue'),
            schema_fields=(('venue_id', 'STRING', 'REQUIRED'), ('name', 'STRING', 'REQUIRED'))
        )

        bigquery_client = bigquery.Client(project=self.setup.project)
        table_ref = bigquery_client.get_table('{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'venue'))
        self.assertEqual(table_ref.schema, [
            bigquery.SchemaField("venue_id", "STRING", mode="REQUIRED"),
            bigquery.SchemaField("name", "STRING", mode="REQUIRED")
        ])

    def test_create_table_with_schema_file_name(self):

        bq_client = bq.Client(project=self.setup.project)
        bq_client.create_table(
            table='{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'venue2'),
            schema_file_name='tests/data/schema.csv'
        )

        bigquery_client = bigquery.Client(project=self.setup.project)
        table_ref = bigquery_client.get_table('{}.{}.{}'.format(self.setup.project, self.setup.dataset_id, 'venue2'))
        self.assertEqual(table_ref.schema, [
            bigquery.SchemaField("venue_id", "STRING", mode="REQUIRED"),
            bigquery.SchemaField("name", "STRING", mode="REQUIRED"),
            bigquery.SchemaField("address", "STRING", mode="REQUIRED")

        ])



from google.cloud import storage


class Client:
    """
    Client to bundle Google Storage functionality.

    Args:
        project (str): The Project ID for the project which the client acts on behalf of.
    """

    def __init__(self, project):
        self.project = project
        self.storage_client = storage.Client(project=self.project)

    def download(self, gs_uri, destination_file_name):
        """Download from Google Storage to local.

        Args:
            gs_uri (str):  The Google Storage uri. For example: ``gs://my_bucket_name/my_filename``.
            destination_file_name (str):  The destination file name. For example: ``/some_path/some_file_name``.
        """

        with open(destination_file_name, 'wb') as file_obj:
            self.storage_client.download_blob_to_file(gs_uri, file_obj)

    def upload(self, source_file_name, bucket_name, blob_name):
        """Upload from local to Google Storage.

        Args:
            source_file_name (str):  The source file name.
            bucket_name (str):  The Google Storage bucket name.
            blob_name (str): The destination file name in the bucket
        """

        bucket = self.storage_client.bucket(bucket_name)
        blob = bucket.blob(blob_name)
        blob.upload_from_filename(source_file_name)

    def create(self, bucket_name):
        """create a bucket in Google Storage.

        Args:
            bucket_name (str):  The Google Storage bucket name.
        """

        self.storage_client.create_bucket(bucket_name, location='EU')






import parse
import paramiko
import os
from tmg.data import logs
import stat


class Client:
    """
    Creates a client which manages the connection to the FTP server.

    Args:
        connection_string (str): The FTP connection string in the format {username}:{password}@{host}:{port}
    """

    def __init__(self, connection_string):
        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        parsed_connection = parse.parse('{username}:{password}@{host}:{port}', connection_string)

        ssh_client.connect(hostname=parsed_connection['host'],
                            username=parsed_connection['username'],
                            password=parsed_connection['password'],
                            port=int(parsed_connection['port']))
        self.connection = ssh_client.open_sftp()

    def list_files(self, dir_path='.'):
        """Lists the files in a directory on the FTP server

        Args:
            dir_path (str): the path of the directory to list the contents of

        Returns:
            list: The list of the contents
        """

        return self.connection.listdir(dir_path)

    def upload_file(self, local_path, remote_path='.'):
        """Uploads a file from the local to the sFTP. If the remote path provided is a directory the filename
        of the local file will be used.

        Args:
            local_path (str): local path to the file to load
            remote_path (str): path to load the file to on the sFTP

        Returns:
            str: The path to the file loaded onto the sFTP
        """

        try:
            fileattr = self.connection.lstat(remote_path)
            if stat.S_ISDIR(fileattr.st_mode):
                filename = os.path.basename(local_path)
                remote_path = os.path.join(remote_path, filename)
        except IOError:
            pass

        self.connection.put(localpath=local_path, remotepath=remote_path)

        return remote_path

    def download_file(self, remote_path, local_path='.'):
        """Downloads a file from the sFTP to the local system. If the local path is a directory the filename of the
        remote file is used.

        Args:
            remote_path (str): The path to the file to download.
            local_path (str): The path to download the file to on the local system

        returns:
            str: The local path the file has been place.
        """

        if os.path.isdir(local_path):
            filename = os.path.basename(remote_path)
            local_path = os.path.join(local_path, filename)

        logs.client.logger.info("Starting file download.")

        self.connection.get(remotepath=remote_path, localpath=local_path)

        logs.client.logger.info("Finished downloading file.")

        return local_path

    def change_file_permission(self, remote_path, permission=stat.S_IRWXU):
        """Changes the permissions of the file on the sFTP

        Args:
            remote_path (str): The path to the file to change to permissions on
            permission (int): The permissions to set the file to
        """

        self.connection.chmod(path=remote_path, mode=permission)

    def delete_file(self, remote_path):
        """Deletes the specified file off the sFTP server

        Args:
            remote_path: path to the file to delete
        """

        self.connection.remove(remote_path)

    def close(self):
        """Closes the connection to the ftp server
        """
        self.connection.close()



import unittest
import csv
from mysql import connector

from tmg.data import mysql
from tests.setup import setup


def setUpModule():
    setup.create_mysql_db()
    setup.create_mysql_table()


def tearDownModule():
    setup.delete_mysql_db()


class TestMySQL(unittest.TestCase):

    def __init__(self, *args, **kwargs):
        super(TestMySQL, self).__init__(*args, **kwargs)
        self.setup = setup

    def test_upload_table(self):

        # create uploaded_actors MySQL table
        connection = connector.connect(
            host=self.setup.mysql_ip_address,
            user=self.setup.mysql_username,
            password=self.setup.mysql_password,
            database=self.setup.mysql_database
        )
        create_table_query = """
            CREATE TABLE uploaded_actors (
                profile_id int,
                first_name varchar(255),
                last_name varchar(255)
            )
        """
        connection.autocommit = True
        cursor = connection.cursor()
        cursor.execute(create_table_query)

        # remove the first row from sample file to make it ready to upload to MySQL
        with open('tests/data/sample.csv') as input_file_obj:
            with open('sample_to_load_to_mysql.csv', 'w+') as output_file_obj:
                for line in input_file_obj:
                    output_file_obj.write(line)

        # run upload
        mysql_client = mysql.Client(
            connection_string='{username}:{password}@{host}:{port}/{database}'.format(
                username=self.setup.mysql_username,
                password=self.setup.mysql_password,
                host=self.setup.mysql_ip_address,
                port=self.setup.mysql_port,
                database=self.setup.mysql_database
            )
        )
        mysql_client.upload_table(
            file_path='sample_to_load_to_mysql.csv',
            database=self.setup.mysql_database,
            table='uploaded_actors'
        )

        # creating list based on destination bucket file values
        with open('tests/data/sample.csv', newline='') as csvfile:
            reader = csv.DictReader(csvfile)
            storage_keys = []
            for row in reader:
                storage_keys.append('{}{}{}'.format(
                    row['profile_id'],
                    row['first_name'] if row['first_name'] else 'NULL',
                    row['last_name'] if row['last_name'] else 'NULL'
                ))

        # creating list based on mysql uploaded table values
        cursor.execute('SELECT profile_id, first_name, last_name from uploaded_actors')
        mysql_keys = []
        for row in cursor.fetchall():
            mysql_keys.append('{}{}{}'.format(
                row[0],
                'NULL' if row[1] is None else row[1],
                'NULL' if row[2] is None else row[2]
            ))
        cursor.close()
        connection.close()

        self.assertEqual(storage_keys, mysql_keys)



import os
import io
import setuptools


name = "tmg-data"
description = "TMG data library"
version = "0.0.5"
dependencies = [
    "bcrypt==3.1.7",
    "cachetools==4.1.0",
    "certifi==2020.4.5.1",
    "cffi==1.14.0",
    "chardet==3.0.4",
    "cryptography==2.9.2",
    "delegator.py==0.1.1",
    "dnspython==1.16.0",
    "google-api-core==1.17.0",
    "google-api-python-client==1.8.2",
    "google-auth==1.14.0",
    "google-auth-httplib2==0.0.3",
    "google-cloud-bigquery==1.24.0",
    "google-cloud-core==1.3.0",
    "google-cloud-storage==1.27.0",
    "google-resumable-media==0.5.0",
    "googleapis-common-protos==1.51.0",
    "httplib2==0.17.3",
    "idna==2.9",
    "Jinja2==2.11.2",
    "MarkupSafe==1.1.1",
    "mysql-connector==2.2.9",
    "oauth2client==4.1.3",
    "paramiko==2.7.1",
    "parse==1.15.0",
    "pexpect==4.8.0",
    "protobuf==3.6.1",
    "ptyprocess==0.6.0",
    "pyasn1==0.4.8",
    "pyasn1-modules==0.2.8",
    "pycparser==2.20",
    "PyNaCl==1.4.0",
    "pysftp==0.2.9",
    "pytz==2019.3",
    "requests==2.23.0",
    "rsa==4.0",
    "six==1.14.0",
    "uritemplate==3.0.1",
    "urllib3==1.25.9"
]

package_root = os.path.abspath(os.path.dirname(__file__))

readme_filename = os.path.join(package_root, "README.rst")
with io.open(readme_filename, encoding="utf-8") as readme_file:
    readme = readme_file.read()


setuptools.setup(
    name=name,
    version=version,
    description=description,
    long_description=readme,
    author='TMG Data Platform team',
    author_email="data.platform@telegraph.co.uk",
    license="Apache 2.0",
    url='https://github.com/telegraph/tmg-data',
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    install_requires=dependencies,
    python_requires='>=3.6',

)



from threading import Thread


class ParallelRunner:
    """
    A runner which runs various transfer operations in parallel.
    """

    def __init__(self):
        self.processes = []

    def add(self, function, *args, **kwargs):
        """
        Adds any function to the list of functions to be run in parallel

        Args:
            function (obj: function): the function to be run in parallel
            args: Any arguments which need to be passed to the function at runtime
            kwargs: Any key word arguments which need to be passed to the function at runtime

        Example:
            >>> from tmg.data import runner
            >>> parallel_runner = runner.ParallelRunner()
            >>> parallel_runner.add(sample_function, 'sample.csv', sample_keyword=True)
        """
        self.processes.append((function, list(args), kwargs))

    def run(self, wait_for_result=False):
        """
        Runs the list of processes in this object in parallel.

        Args:
            wait_for_result (bool, Optional): Determines whether to wait for all the processes to finish before returning from this function

        Example:
            >>> from tmg.data import runner
            >>> parallel_runner = runner.ParallelRunner()
            >>> parallel_runner.add(sample_function, 'sample.csv', sample_keyword=True)
            >>> parallel_runner.run(wait_for_result=True)
        """
        threads = [ThreadWithReturnValue(target=process[0], args=process[1], kwargs=process[2]) for process in self.processes]
        for thread in threads:
            thread.start()
            thread.run()
        if wait_for_result:
            results = []
            for thread in threads:
                results.append(thread.join())
            return results


class ThreadWithReturnValue(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, *, daemon=None):
        Thread.__init__(self, group, target, name, args, kwargs, daemon=daemon)

        self._return = None

    def run(self):
        if self._target is not None:
            self._return = self._target(*self._args, **self._kwargs)

    def join(self):
        Thread.join(self)
        return self._return
